import os
import sqlite3
import time
import typing

from sqlite2pg.modules import Logger

__all__: typing.List[str] = [
    "Worker",
]


# Represents the schema dicts used later on.
RawSchemaT = typing.MutableMapping[str, typing.List[typing.Sequence[str]]]
CleanSchemaT = typing.MutableMapping[str, typing.List[str]]


class Worker:
    """The object that carries the workload of the program.

    Args:
        logger: sqlite2pg.Logger
            The logger to use for the worker. This logger will be
            supplied automatically by the command line entry point.
    """

    __slots__: typing.Sequence[str] = ("log", "test_sqlite_db")

    def __init__(self, logger: Logger) -> None:
        self.log = logger
        self.log.debug("worker initialized...")
        self.test_sqlite_db = "./database.db3"

    def get_sqlite_schema(self, db: str) -> CleanSchemaT:
        """Gets the schema for a given sqlite3 database.3

        Args:
            db: str
                The path to the sqlite database to get the schema for.

        Returns:
            CleanSchemaT:
                A mapping containing the schema for the tables found
                in the database.
        """

        start: float = time.time()

        # Validate that the database file exists
        if not os.path.isfile(db):
            self.log.error(f"can't connect to '{db}', the file doesnt exist.")
            exit(1)

        # Connect to the database, and obtain a cursor.
        self.log.debug(f"connecting to: '{db}'.")
        conn: sqlite3.Connection = sqlite3.connect(db)
        cur: sqlite3.Cursor = conn.cursor()

        # Assign 2 mappings for raw and clean schema.
        raw_schema: RawSchemaT = {}
        clean_schema: CleanSchemaT = {}

        try:
            self.log.debug("attempting to fetch tables from sqlite master.")

            # Gather all table names
            cur.execute(
                """SELECT name FROM sqlite_master
                WHERE type IS 'table'
                AND name NOT LIKE 'sqlite_%'
                ORDER BY 1
                """
            )

        # We were unable to fetch the data, likely it was not a database.
        except sqlite3.DatabaseError as e:
            self.log.debug(f"{e.__class__.__name__}: {e}")
            self.log.error(f"'{db}' is not a database. exiting...")
            exit(1)

        else:
            self.log.debug("connection secured. continuing...")

        # Iterate through table names.
        for table in [d[0] for d in cur.fetchall()]:
            self.log.info(f"fetching schema for '{table}'.")

            # Gather the schema for that table.
            cur.execute(
                f"""SELECT sql
                FROM sqlite_master
                WHERE name='{table}'
                """
            )

            # If there is no schema, error.
            if not (data := cur.fetchall()):
                self.log.error(f"found '{table}' but no schema. exiting...")
                exit(1)

            # Assign that schema to our schema dict.
            self.log.debug(f"found schema for '{table}'.")
            raw_schema[table] = data

        # We didn't find tables or schema.
        if not raw_schema:
            self.log.error("no tables found. exiting...")
            exit(1)

        # Replace tabs with 4 spaces and create
        # a list of strings instead of a list
        # of tuples with one string inside.
        # Add this list to the clean mapping.
        for t, q in raw_schema.items():
            clean_schema[t] = [q[0][0].replace("\t", "    ")]

        # Close the connection to sqlite
        self.log.debug(f"closed connection to sqlite: '{db}'.")
        conn.close()

        end: float = time.time()

        for s in clean_schema.values():
            print(f"{s[0]}\n")

        self.log.info(
            f"found {len(clean_schema)} tables and schema in {(end - start) * 1000:.4f} ms."
        )

        return clean_schema

    def convert_sqlite_to_pg(self, schema: CleanSchemaT) -> CleanSchemaT:
        """Converts sqlite queries to postres syntax.

        Args:
            schema: CleanSchemaT
                A mutable mapping of tables to table schema.

        Returns:
            CleanSchemaT:
                A mapping containing the schema for the tables from
                sqlite transformed to fit postgres syntax.
        """

        start: float = time.time()
        self.log.debug("beginning conversion to postgres syntax")

        # Iterate over each table and alter its schema
        for table, query in schema.items():
            buffer: str = query[0]

            # Postgres doesn't like double quotes.
            buffer = buffer.replace('"', "'")
            # Sqlite integers can be larger than pgsql.
            # Handle upper and lower case.
            buffer = buffer.replace("integer", "bigint")
            buffer = buffer.replace("INTEGER", "BIGINT")

            # Append a semicolon to the query, if
            # there is not already one there.
            if buffer[-1] != ";":
                buffer += ";"

            # Assign the new schema back to the table
            # now the that is has been converted.
            query[0] = buffer
            self.log.debug(f"'{table}' has been converted.")

        end: float = time.time()

        print(
            f"converted {len(schema)} tables in {(end - start) * 1000:.4f} ms. Obtained new schema."
        )

        for s in schema.values():
            print(f"\n{s[0]}")

        return schema

    def validate_input(self, message: str, bad_actor: int = 0) -> str:
        """Validates user confirmation input.

        Args:
            message: str
                The message received from stdin.
            bad_actor: int, optional
                The amount of invalid inputs received thus far.
                Defaults to 0.

        Returns:
            str: The text from the users confirmation
        """

        self.log.debug(f"Input validation (attempt {bad_actor + 1}).")
        validated: str = input(message)

        # Handle case where user inputs nothing.
        while not validated:
            bad_actor += 1

            if bad_actor > 3:
                self.log.error("Too many failed inputs. Exiting...")
                exit(1)

            validated = self.validate_input(message, bad_actor)

        if validated[0] not in "yY":
            self.log.debug(f"User chose to exit with input: '{validated}'.")
            self.log.info("Exiting...")
            exit(0)

        self.log.debug(f"User accepted with input: '{validated}'.")
        return validated

    def execute(self) -> None:
        """Execute the main logic of the program."""

        print("<-- v**v -| SQLITE2PG |-- v**v -->")

        # Gather the sqlite schema from the database.
        sqlite_schema = self.get_sqlite_schema(self.test_sqlite_db)

        # Validate we gather the correct schema.
        self.validate_input("does the schema look correct [y/n]: ")

        # Validate the converted schema is correct
        converted_schema = self.convert_sqlite_to_pg(sqlite_schema)

        # Validate the converted schema is correct
        self.validate_input("does the converted schema look correct [y/n]: ")

        # TODO
        # Connect to postgres
        # Create the tables
        # Get the actual data from the tables in sqlite
        # Insert the data into postgres
        # Could we use pandas to move the data?



import logging
import time
import typing


__all__: typing.List[str] = [
    "Logger",
]


DEFAULT_LOG_FILE: str = f"./slqlite2pg-{time.time():.0f}.log"


class Logger(logging.Logger):
    """Logging for the program."""

    __slots__: typing.Sequence[str] = ()

    @staticmethod
    def new(
        stream_level: int = logging.INFO,
        file_level: int = logging.DEBUG,
        file_path: str = None,
    ) -> logging.Logger:
        """Initializes a new logger object and returns it.

        Args:
            stream_level: int
                The logging level to set for the stream handler. This will
                be the output to the terminal. Defaults to INFO.

            file_level: int
                The logging level to set for the stream handler. This will
                be the output to the terminal. Defaults to DEBUG.

            file_path: str
                The absolute or relative path you would like the log file
                to generated at. Defaults to './sqlite2pg-{%epoch%}.log'
                where {%epoch%} is the current epoch time.

        Returns:
            logging.Logger: The configured logger.
        """

        # Instantiate logger.
        _log: logging.Logger = logging.getLogger()
        _log.setLevel(logging.DEBUG)

        # Instantiate formatter.
        _logF: logging.Formatter = logging.Formatter(
            f"[%(asctime)s] %(levelname)8s ||| %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )

        # Instantiate streamhandler (stdout) and set log level.
        _logSh: logging.StreamHandler = logging.StreamHandler()
        _logSh.setLevel(stream_level)

        # Instantiate file handler.
        _logFh: logging.FileHandler = logging.FileHandler(
            file_path if file_path else DEFAULT_LOG_FILE
        )

        # Set file handler formatter, and log level.
        _logFh.setFormatter(_logF)
        _logFh.setLevel(file_level)

        # Add both handlers and return the Logger object.
        _log.addHandler(_logFh)
        _log.addHandler(_logSh)

        return _log



import logging

from sqlite2pg import Logger
from sqlite2pg import Worker


logger: logging.Logger = Logger.new(
    file_path="./test.log",
)


def main():
    """Entry point for the program."""

    logger.debug("sqlite2pg starting...")

    worker = Worker(logger)
    worker.execute()

    logger.debug("cleaned up. exiting...")


if __name__ == "__main__":
    main()



import abc
import typing


__all__: typing.List[str] = [
    "Sqlite2pgError",
    "SqliteError",
    "PostgresError",
]


# Do we need these???


class Sqlite2pgError(Exception):
    """Base Exception that all sqlite2pg exceptions inherit from."""

    __slots__: typing.Sequence[str] = ()
    pass


class SqliteError(Sqlite2pgError):
    """Represents an error in an sqlite3 related action."""

    __slots__: typing.Sequence[str] = ()
    pass


class PostgresError(Sqlite2pgError):
    """Represents an error in an sqlite3 related action."""

    __slots__: typing.Sequence[str] = ()
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sqlite2pg', 'sqlite2pg.modules']

package_data = \
{'': ['*']}

install_requires = \
['aiosqlite>=0.17.0,<0.18.0', 'asyncpg>=0.24.0,<0.25.0']

entry_points = \
{'console_scripts': ['sqlite2pg = sqlite2pg.cli:main']}

setup_kwargs = {
    'name': 'sqlite2pg',
    'version': '0.1.0',
    'description': 'An Sqlite3 to Postgresql database migration tool.',
    'long_description': '# sqlite2pg\n\nAn Sqlite3 to Postgresql database migration tool.\n\n## A warning\n\nThis project is still in very early development, and will not be ready until the v1.0 release.\nPlease refrain from using sqlite2pg for now.\n\n## Why sqlite2pg\n\n - An easy to use command line interface.\n - Options to accomodate different use cases.\n - Complete full migrations, or just generate schema.\n - Asynchronous, keeping performance in mind for big jobs.\n\n## Installation\n\nsqlite2pg requires python 3.8.0 or greater.\n\nTo get started:\n```bash\npip install sqlite2pg\n```\n\n## License\n\nsqlite2pg is licensed under the [BSD 3-Clause License](https://github.com/Jonxslays/sqlite2pg/blob/master/LICENSE).\n',
    'author': 'Jonxslays',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Jonxslays/sqlite2pg',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<=3.10',
}


setup(**setup_kwargs)



import typing

from .config import *
from .errors import *
from .worker import *

__all__: typing.List[str] = [
    "Worker",
    "Logger",
    "Sqlite2pgError",
    "SqliteError",
    "PostgresError",
]
